
/* BWT.C */

#include <stdio.h>
#include <stdlib.h>

#include "bwt.h"
#include "types.h"
#include "algo.h"
#include "bitio.h"
#include "debug.h"
#include "hufblock.h"
#include "mtf.h"
#include "rle.h"

/* Variables */

int    Index, *P;
uchar *L, **M;

static int    C[256];
static uchar *F;

/* Prototypes */

static void AllocBWTBlock (void);
static void FreeBWTBlock  (void);

static void AllocMatrix   (void);
static void FreeMatrix    (void);
static void AllocArray    (void);
static void FreeArray     (void);

static void SortBlock     (void);
static void CompLast      (void);

static void CompFirst     (void);
static void CompBlock     (void);

static int  StringCmp     (const void *String1, const void *String2);

/* Functions */

/* AllocBWTBlock() */

static void AllocBWTBlock(void) {

   F = malloc(2*N);
   if (F == NULL) FatalError("AllocBWTBlock(): Not enough memory");
   L = &F[N];
}

/* FreeBWTBlock() */

static void FreeBWTBlock(void) {

   if (F != NULL) {
      free(F);
      F = NULL;
      L = NULL;
   }
}

/* CodeBWT() */

void CodeBWT(void) {

   AllocBWTBlock();
   AllocMatrix();
   SortBlock();
   CompLast();
   FreeMatrix();

   if (Verbosity >= 2) fprintf(stderr,"I = %d\n",Index);

   SendBits(25,Index);

   CodeMTF(L,N);

   AllocHufBlock();
   CodeRLE();
   SendHufBlock();
   FreeHufBlock();
   FreeBWTBlock();
}

/* DecodeBWT() */

void DecodeBWT(void) {

   Index = GetBits(25);
   if (Verbosity >= 2) fprintf(stderr,"I = %d\n",Index);

   AllocBWTBlock();
   AllocHufBlock();
   GetHufBlock();
   DecodeRLE();
   FreeHufBlock();

   DecodeMTF(L,N);

   AllocArray();
   CompFirst();
   CompBlock();
   FreeArray();
   FreeBWTBlock();
}

/* AllocMatrix() */

static void AllocMatrix(void) {

   M = malloc((size_t)(N*sizeof(uchar *)));
   if (M == NULL) FatalError("AllocMatrix(): Not enough memory");
}

/* FreeMatrix() */

static void FreeMatrix(void) {

   if (M != NULL) {
      free(M);
      M = NULL;
   }
}

/* AllocArray() */

static void AllocArray(void) {

   P = malloc((size_t)(N*sizeof(int)));
   if (P == NULL) FatalError("AllocArray(): Not enough memory");
}

/* FreeArray() */

static void FreeArray(void) {

   if (P != NULL) {
      free(P);
      P = NULL;
   }
}

/* SortBlock() */

static void SortBlock(void) {

   int I;

   L = &F[N];

   for (I = 0; I < N; I++) {
      F[I] = L[I] = S[I];
      M[I] = &F[I];
   }

   qsort(M,(size_t)N,sizeof(uchar*),&StringCmp);
}

/* StringCmp() */

static int StringCmp(const void *String1, const void *String2) {

   int I, Diff;
   uchar *C1, *C2;

   C1 = *((uchar **) String1);
   C2 = *((uchar **) String2);

   I = N;
   do {
      Diff = *C1 - *C2;
      if (Diff != 0) return Diff;
      C1++;
      C2++;
   } while (--I > 0);

   return 0;
}

/* CompLast() */

static void CompLast(void) {

   int I;
   uchar *String;

   for (I = 0; I < N; I++) {
      if (M[I] == F) Index = I;
      String = &M[I][N-1];
      if (String >= L) String -= N;
      L[I] = *String;
   }
}

/* CompFirst() */

static void CompFirst(void) {

   int I, Char, Sum, SumOld;

   for (Char = 0; Char < 256; Char++) C[Char] = 0;

   for (I = 0; I < N; I++) {
      Char = L[I];
      P[I] = C[Char]++;
   }

   Sum = 0;
   for (Char = 0; Char < 256; Char++) {
      SumOld = Sum;
      Sum += C[Char];
      C[Char] = SumOld;
   }
}

/* CompBlock() */

static void CompBlock(void) {

   int I, J, Char;

   I = Index;
   for (J = N-1; J >= 0; J--) {
      Char = L[I];
      S[J] = L[I];
      I = C[Char] + P[I];
   }
}

/* End of BWT.C */

